#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""This module has the input data for heat_driver UTs."""


class DummyDictionaries(object):
    """Implements the input data for heat_driver UTs.

    This class holds the input data that are required in
    testing the heat_driver test cases.
    """

    DEFAULT_LBV2_CONFIG = {
        "heat_template_version": "2015-10-15",
        "description": "Configuration for Haproxy Neutron LB V2 service",
        "parameters": {
            "lb_port": {
                "type": "number",
                "default": 80,
                "description": "Port used by the load balancer"
            },
            "app_port": {
                "type": "number",
                "default": 80,
                "description": "Port used by the servers"
            },
            "Subnet": {
                "type": "string",
                "description": "Subnet on which the LB will be located"
            },
            "vip_ip": {
                "type": "string",
                "description": "VIP IP Address"
            },
            "service_chain_metadata": {
                "type": "string",
                "description": "sc metadata"
            }
        },
        "resources": {
            "monitor": {
                "type": "OS::Neutron::LBaaS::HealthMonitor",
                "properties": {
                    "delay": 3,
                    "type": "HTTP",
                    "timeout": 3,
                    "max_retries": 3,
                    "pool": {
                        "get_resource": "pool"
                    }
                }
            },
            "pool": {
                "type": "OS::Neutron::LBaaS::Pool",
                "properties": {
                    "lb_algorithm": "ROUND_ROBIN",
                    "protocol": "HTTP",
                    "listener": {
                        "get_resource": "listener"
                    }
                }
            },
            "listener": {
                "type": "OS::Neutron::LBaaS::Listener",
                "properties": {
                    "loadbalancer": {
                        "get_resource": "loadbalancer"
                    },
                    "protocol": "HTTP",
                    "protocol_port": {
                        "get_param": "lb_port"
                    }
                }
            },
            "loadbalancer": {
                "type": "OS::Neutron::LBaaS::LoadBalancer",
                "properties": {
                    "vip_subnet": {
                        "get_param": "Subnet"
                    },
                    "provider": "loadbalancerv2",
                    "vip_address": {
                        "get_param": "vip_ip"
                    },
                    "description": {
                        "get_param": "service_chain_metadata"
                    }
                }
            }
        }
    }

    DEFAULT_FW_CONFIG = {
        'heat_template_version': '2013-05-23',
        'description': 'Template to deploy firewall',
        'resources': {
            'sc_firewall_rule3': {
                'type': 'OS::Neutron::FirewallRule',
                'properties': {
                    'action': 'allow',
                    'destination_port': '82',
                    'protocol': 'tcp', 'name': 'Rule_3'
                }
            },
            'sc_firewall_rule2': {
                'type': 'OS::Neutron::FirewallRule',
                'properties': {
                    'action': 'allow',
                    'destination_port': '81',
                    'protocol': 'tcp', 'name': 'Rule_2'
                }
            },
            'sc_firewall_rule1': {
                'type': 'OS::Neutron::FirewallRule',
                'properties': {
                    'action': 'allow',
                    'destination_port': '80',
                    'protocol': 'tcp',
                    'name': 'Rule_1'
                }
            },
            'sc_firewall_rule0': {
                'type': 'OS::Neutron::FirewallRule',
                'properties': {
                    'action': 'allow',
                    'destination_port': '22',
                    'protocol': 'tcp', 'name': 'Rule_0'
                }
            },
            'sc_firewall_rule4': {
                'type': 'OS::Neutron::FirewallRule',
                'properties': {
                    'action': 'allow',
                    'protocol': 'icmp',
                    'name': 'Rule_4'
                }
            },
            'sc_firewall_policy': {
                'type': 'OS::Neutron::FirewallPolicy',
                'properties': {
                    'name': '',
                    'firewall_rules': [
                         {'get_resource': 'sc_firewall_rule0'},
                         {'get_resource': 'sc_firewall_rule1'},
                         {'get_resource': 'sc_firewall_rule2'},
                         {'get_resource': 'sc_firewall_rule3'},
                         {'get_resource': 'sc_firewall_rule4'}]
                }
            },
            'sc_firewall': {
                'type': 'OS::Neutron::Firewall',
                'properties': {
                    'firewall_policy_id': {
                         'get_resource': 'sc_firewall_policy'
                    },
                    'name': 'serviceVM_infra_FW',
                    'description': {'insert_type': 'east_west'}
                }
            }
        }
    }

    DEFAULT_VPN_CONFIG = {
        'resources': {
            'IKEPolicy': {
                'type': 'OS::Neutron::IKEPolicy',
                'properties': {
                    'name': 'IKEPolicy',
                    'auth_algorithm': 'sha1',
                    'encryption_algorithm': '3des',
                    'pfs': 'group5',
                    'lifetime': {
                        'units': 'seconds',
                        'value': 3600
                    },
                    'ike_version': 'v1',
                    'phase1_negotiation_mode': 'main'
                }
            },
            'VPNService': {
                'type': 'OS::Neutron::VPNService',
                'properties': {
                    'router_id': {
                        'get_param': 'RouterId'
                    },
                    'subnet_id': {
                        'get_param': 'Subnet'
                    },
                    'admin_state_up': 'true',
                    'description': {
                        'get_param': 'ServiceDescription'
                    },
                    'name': 'VPNService'
                }
            },
            'site_to_site_connection1': {
                'type': 'OS::Neutron::IPsecSiteConnection',
                'properties': {
                    'psk': 'secret',
                    'initiator': 'bi-directional',
                    'name': 'site_to_site_connection1',
                    'admin_state_up': 'true',
                    'description':
                        'fip=1.103.1.20;tunnel_local_cidr=11.0.1.0/24;\
                        user_access_ip=1.103.2.20;fixed_ip=192.168.0.3;\
                        standby_fip=1.103.1.21;service_vendor=vyos;\
                        stitching_cidr=192.168.0.0/28;\
                        stitching_gateway=192.168.0.1;mgmt_gw_ip=120.0.0.1',
                    'peer_cidrs': ['11.0.0.0/24'],
                    'mtu': 1500,
                    'ikepolicy_id': {
                        'get_resource': 'IKEPolicy'
                    },
                    'dpd': {
                        'interval': 30,
                        'actions': 'hold',
                        'timeout': 120
                    },
                    'vpnservice_id': {
                        'get_resource': 'VPNService'
                    },
                    'peer_address': '1.103.2.88',
                    'peer_id': '1.103.2.88',
                    'ipsecpolicy_id': {
                        'get_resource': 'IPsecPolicy'
                    }
                }
            },
            'IPsecPolicy': {
                'type': 'OS::Neutron::IPsecPolicy',
                'properties': {
                    'name': 'IPsecPolicy',
                    'transform_protocol': 'esp',
                    'auth_algorithm': 'sha1',
                    'encapsulation_mode': 'tunnel',
                    'encryption_algorithm': '3des',
                    'pfs': 'group5',
                    'lifetime': {
                        'units': 'seconds',
                        'value': 3600
                    }
                }
            }
        }
    }

    appended_sc_firewall_policy = {
        'type': 'OS::Neutron::FirewallPolicy',
        'properties': {
            'name': '',
            'firewall_rules': [
                {
                    'get_resource': 'sc_firewall_rule0'
                },
                {'get_resource': 'sc_firewall_rule1'},
                {'get_resource': 'sc_firewall_rule2'},
                {'get_resource': 'sc_firewall_rule3'},
                {'get_resource': 'sc_firewall_rule4'},
                {'get_resource': 'node_driver_rule_2b86019a-45f7-44_1'},
                {'get_resource': 'node_driver_rule_2b86019a-45f7-44_2'},
                {'get_resource': 'node_driver_rule_2b86019a-45f7-44_3'},
                {'get_resource': 'node_driver_rule_2b86019a-45f7-44_4'},
                {'get_resource': 'node_driver_rule_2b86019a-45f7-44_5'},
            ]
        }
    }

    updated_sc_firewall_policy = {
        'type': 'OS::Neutron::FirewallPolicy',
        'properties': {
            'name': '-fw_redirect',
            'firewall_rules': [
                {'get_resource': 'node_driver_rule_af6a8a58-1e25-49_1'},
                {'get_resource': 'node_driver_rule_af6a8a58-1e25-49_2'},
                {'get_resource': 'node_driver_rule_af6a8a58-1e25-49_3'},
                {'get_resource': 'node_driver_rule_af6a8a58-1e25-49_4'},
                {'get_resource': 'node_driver_rule_af6a8a58-1e25-49_5'},
            ]
        }
    }

    updated_template_sc_firewall_policy = {
        'type': 'OS::Neutron::FirewallPolicy',
        'properties': {
            'name': '',
            'firewall_rules': [
                {'get_resource': 'node_driver_rule_af6a8a58-1e25-49_1'},
                {'get_resource': 'node_driver_rule_af6a8a58-1e25-49_2'},
                {'get_resource': 'node_driver_rule_af6a8a58-1e25-49_3'},
                {'get_resource': 'node_driver_rule_af6a8a58-1e25-49_4'},
                {'get_resource': 'node_driver_rule_af6a8a58-1e25-49_5'},
            ]
        }
    }

    policy_targets = {
        'policy_targets': [
            {'name': 'provider_0132c_00b93',
             'port_id': 'dde7d849-4c7c-4b48-8c21-f3f52c646fbe',
             'id': "dde7d849-4c7c-4b48-8c21-f3f52c646fbf",
             'policy_target_group_id': "dde7d849-4c7c-4b48-8c21-f3f52c646fbg"}]
    }

    policy_target = {
        'policy_target': {
            'name': 'service_target_provider_0132c_00b93'
        }
    }

    port_info = {
        'port': {
            'status': 'ACTIVE',
            'binding:host_id': 'LibertyCompute',
            'name': '',
            'allowed_address_pairs': [],
            'admin_state_up': True,
            'network_id': '2286b432-a443-4cd3-be49-e354f531abe3',
            'dns_name': '',
            'extra_dhcp_opts': [],
            'mac_address': 'fa:16:3e:43:34:33',
            'dns_assignment': [
                {'hostname': 'host-42-0-0-13',
                 'ip_address': '42.0.0.13',
                 'fqdn': 'host-42-0-0-13.openstacklocal.'
                 }],
            'binding:vif_details': {
                'port_filter': True,
                'ovs_hybrid_plug': True
            },
            'binding:vif_type': 'ovs',
            'device_owner': 'compute:nova',
            'tenant_id': 'f6b09b7a590642d8ac6de73df0ab0686',
            'binding:profile': {},
            'binding:vnic_type': 'normal',
            'fixed_ips': [
                {'subnet_id': 'b31cdafe-bdf3-4c19-b768-34d623d77d6c',
                 'ip_address': '42.0.0.13'}],
            'id': 'dde7d849-4c7c-4b48-8c21-f3f52c646fbe',
            'security_groups': ['ad3b95a4-b5ce-4a95-9add-6ef2ee797e72'],
            'device_id': '36e9a6d9-ea04-4627-93c5-6f708368c070'
        }
    }
    provider_ptg = {
        'shared': False,
        'subnets': ['a2702d68-6deb-425c-a266-e27b349e00ce'],
        'proxy_group_id': None,
        'description': '',
        'consumed_policy_rule_sets': [],
        'network_service_policy_id': '0cdf2cba-90f8-44da-84a5-876e582f6e35',
        'tenant_id': '8ae6701128994ab281dde6b92207bb19',
        'service_management': False,
        'provided_policy_rule_sets': ['7d4b1ef2-eb80-415d-ad13-abf0ea0c52f3'],
        'policy_targets': [
            {'name': 'provider_0132c_00b93',
             'port_id': 'dde7d849-4c7c-4b48-8c21-f3f52c646fbe'}],
        'proxy_type': None,
        'proxied_group_id': None,
        'l2_policy_id': '120aa972-1b58-418d-aa5b-1d2f96612c49',
        'id': 'af6a8a58-1e25-49c4-97a3-d5f50b3aa04b',
        'name': 'fw_redirect'
    }

    consumer_ptg = {
        'shared': False,
        'subnets': ['a2702d68-6deb-425c-a266-e27b349e00ce'],
        'proxy_group_id': None,
        'description': '',
        'consumed_policy_rule_sets': ['7d4b1ef2-eb80-415d-ad13-abf0ea0c52f3'],
        'network_service_policy_id': '0cdf2cba-90f8-44da-84a5-876e582f6e35',
        'tenant_id': '8ae6701128994ab281dde6b92207bb19',
        'service_management': False,
        'provided_policy_rule_sets': [],
        'policy_targets': [
            {'name': 'provider_0132c_00b93',
             'port_id': 'dde7d849-4c7c-4b48-8c21-f3f52c646fbe'}],
        'proxy_type': None,
        'proxied_group_id': None,
        'l2_policy_id': '120aa972-1b58-418d-aa5b-1d2f96612c49',
        'id': 'af6a8a58-1e25-49c4-97a3-d5f50b3aa04b',
        'name': 'fw_redirect'
    }

    l3_policies = {
        'l3_policies': [
            {'tenant_id': '8ae6701128994ab281dde6b92207bb19',
             'name': 'remote-vpn-client-pool-cidr-l3policy'}]
    }

    policy_rule_sets = {
        'policy_rule_sets': [
            {'id': '7d4b1ef2-eb80-415d-ad13-abf0ea0c52f3',
             'name': 'fw_redirect',
             'policy_rules': ['493788ad-2b9a-47b1-b04d-9096d4057fb5'],
             'tenant_id': '8ae6701128994ab281dde6b92207bb19',
             'shared': False,
             'consuming_policy_target_groups':
             ['af6a8a58-1e25-49c4-97a3-d5f50b3aa04b'],
             'consuming_external_policies': None}]
    }

    policy_rules = {
        'policy_rules': [
            {'id': '493788ad-2b9a-47b1-b04d-9096d4057fb5',
             'name': 'fw_redirect',
             'policy_actions': ['0bab5fa6-4f89-4e15-8363-dacc7d825466'],
             'policy_classifier_id': '8e5fc80f-7544-484c-82d0-2a5794c10664',
             'tenant_id': '8ae6701128994ab281dde6b92207bb19',
             'shared': False}]
    }

    policy_actions = {
        'policy_actions': [
            {'id': '0bab5fa6-4f89-4e15-8363-dacc7d825466',
             'name': 'fw_redirect',
             'action_value': '1e83b288-4b56-4851-83e2-69c4365aa8e5',
             'action_type': 'redirect',
             'tenant_id': '8ae6701128994ab281dde6b92207bb19',
             'shared': False}]
    }

    policy_target_groups = {
        'policy_target_groups': [
            {'shared': False,
             'subnets': ['a2702d68-6deb-425c-a266-e27b349e00ce'],
             'proxy_group_id': None,
             'description': '',
             'consumed_policy_rule_sets': [],
             'network_service_policy_id':
             '0cdf2cba-90f8-44da-84a5-876e582f6e35',
             'tenant_id': '8ae6701128994ab281dde6b92207bb19',
             'service_management': False,
             'provided_policy_rule_sets':
                 ['7d4b1ef2-eb80-415d-ad13-abf0ea0c52f3'],
             'policy_targets': [
                 {'name': 'provider_0132c_00b93',
                  'port_id': 'dde7d849-4c7c-4b48-8c21-f3f52c646fbe'}],
             'proxy_type': None,
             'proxied_group_id': None,
             'l2_policy_id': '120aa972-1b58-418d-aa5b-1d2f96612c49',
             'id': 'af6a8a58-1e25-49c4-97a3-d5f50b3aa04b',
             'name': 'fw_redirect'}]
    }

    subnet_info = {
        'subnet': {
            'name': 'lb-subnet',
            'enable_dhcp': True,
            'network_id': '2286b432-a443-4cd3-be49-e354f531abe3',
            'tenant_id': 'f6b09b7a590642d8ac6de73df0ab0686',
            'dns_nameservers': [],
            'ipv6_ra_mode': None,
            'allocation_pools': [{
                'start': '42.0.0.2', 'end': '42.0.0.254'}],
            'gateway_ip': '42.0.0.1',
            'ipv6_address_mode': None,
            'ip_version': 4,
            'host_routes': [],
            'cidr': '42.0.0.0/24',
            'id': 'b31cdafe-bdf3-4c19-b768-34d623d77d6c',
            'subnetpool_id': None
        }
    }

    subnets_info = {
        'subnets': [
            {'name': 'lb-subnet',
             'enable_dhcp': True,
             'network_id': '2286b432-a443-4cd3-be49-e354f531abe3',
             'tenant_id': 'f6b09b7a590642d8ac6de73df0ab0686',
             'dns_nameservers': [],
             'ipv6_ra_mode': None,
             'allocation_pools': [{
                 'start': '42.0.0.2', 'end': '42.0.0.254'}],
             'gateway_ip': '42.0.0.1',
             'ipv6_address_mode': None,
             'ip_version': 4,
             'host_routes': [],
             'cidr': '42.0.0.0/24',
             'id': 'b31cdafe-bdf3-4c19-b768-34d623d77d6c',
             'subnetpool_id': None}]
    }

    external_policies = {'external_policies': {}}

    fw_template_properties = {
        'fw_rule_keys': ['sc_firewall_rule3', 'sc_firewall_rule2',
                         'sc_firewall_rule1', 'sc_firewall_rule0',
                         'sc_firewall_rule4'],
        'name': '2b8',
        'properties_key': 'properties',
        'resources_key': 'resources',
        'is_template_aws_version': False,
        'fw_policy_key': 'sc_firewall_policy'
    }

    fw_scn_config = "{\"heat_template_version\": \"2013-05-23\",\
        \"description\": \"Template to deploy firewall\", \"resources\":\
        {\"sc_firewall_rule3\": {\"type\": \"OS::Neutron::FirewallRule\",\
        \"properties\": {\"action\": \"allow\", \"destination_port\": \"82\",\
        \"protocol\": \"tcp\", \"name\": \"Rule_3\"}}, \"sc_firewall_rule2\":\
        {\"type\": \"OS::Neutron::FirewallRule\", \"properties\": {\"action\":\
        \"allow\", \"destination_port\": \"81\", \"protocol\": \"tcp\",\
        \"name\": \"Rule_2\"}}, \"sc_firewall_rule1\": {\"type\":\
        \"OS::Neutron::FirewallRule\", \"properties\": {\"action\": \"allow\",\
        \"destination_port\": \"80\", \"protocol\": \"tcp\", \"name\":\
        \"Rule_1\"}}, \"sc_firewall_rule0\": {\"type\":\
        \"OS::Neutron::FirewallRule\", \"properties\": {\"action\": \"allow\",\
        \"destination_port\": \"22\", \"protocol\": \"tcp\", \"name\":\
        \"Rule_0\"}}, \"sc_firewall_rule4\": {\"type\":\
        \"OS::Neutron::FirewallRule\", \"properties\": {\"action\": \"allow\",\
        \"protocol\": \"icmp\", \"name\": \"Rule_4\"}}, \"sc_firewall_policy\"\
        :{\"type\": \"OS::Neutron::FirewallPolicy\", \"properties\": {\"name\"\
        :\"\", \"firewall_rules\": [{\"get_resource\": \"sc_firewall_rule0\"},\
        {\"get_resource\": \"sc_firewall_rule1\"}, {\"get_resource\":\
        \"sc_firewall_rule2\"}, {\"get_resource\": \"sc_firewall_rule3\"},\
        {\"get_resource\": \"sc_firewall_rule4\"}]}}, \"sc_firewall\":\
        {\"type\": \"OS::Neutron::Firewall\", \"properties\":\
        {\"firewall_policy_id\": {\"get_resource\": \"sc_firewall_policy\"},\
        \"description\": \"{\'insert_type\': \'east_west\',\
        \'vm_management_ip\': u'192.168.20.138', \'provider_ptg_info\':\
        [\'fa:16:3e:28:7d:b2\']}\", \"name\": \"serviceVM_infra_FW\"}}}}"

    lbv2_scn_config = "{\"heat_template_version\": \"2015-10-15\",\
        \"description\": \"Configuration for Haproxy Neutron LB V2 service\",\
        \"parameters\": {\"Subnet\": {\"type\": \"string\", \"description\":\
        \"Subnet on which the load balancer will be located\"}, \
        \"service_chain_metadata\": {\"type\": \"string\", \"description\":\
        \"sc metadata\"}, \"vip_ip\": {\"type\": \"string\", \"description\":\
        \"VIP IP Address\"}, \"app_port\": {\"default\": 80, \"type\":\
        \"number\", \"description\": \"Port used by the servers\"}, \
        \"lb_port\": {\"default\": 80, \"type\": \"number\", \"description\":\
        \"Port used by the load balancer\"}}, \"resources\": {\"listener\":\
        {\"type\": \"OS::Neutron::LBaaS::Listener\", \"properties\":\
        {\"protocol_port\": {\"get_param\": \"lb_port\"}, \"protocol\":\
        \"HTTP\", \"loadbalancer\": {\"get_resource\": \"loadbalancer\"}}},\
        \"loadbalancer\": {\"type\": \"OS::Neutron::LBaaS::LoadBalancer\",\
        \"properties\": {\"vip_subnet\": {\"get_param\": \"Subnet\"},\
        \"vip_address\": {\"get_param\": \"vip_ip\"}, \"description\":\
        {\"get_param\": \"service_chain_metadata\"}, \"provider\":\
        \"loadbalancerv2\"}}, \"monitor\": {\"type\":\
        \"OS::Neutron::LBaaS::HealthMonitor\", \"properties\": {\"delay\": 3,\
        \"max_retries\": 3, \"type\": \"HTTP\", \"pool\": {\"get_resource\":\
        \"pool\"}, \"timeout\": 3}}, \"pool\": {\"type\": \
        \"OS::Neutron::LBaaS::Pool\", \"properties\":\
        {\"lb_algorithm\": \"ROUND_ROBIN\", \"listener\": {\"get_resource\":\
        \"listener\"}, \"protocol\": \"HTTP\"}}}}"

    vpn_scn_config = "{\"description\":\"Createsnewvpnservice-ike+ipsec+\
        vpnservice+site-siteconnection(s)\", \"heat_template_version\
        \":\"2013-05-23\", \"parameters\":{\"RouterId\":{\"description\
        \":\"RouterID\", \"type\":\"string\"}, \"ServiceDescription\":{\
        \"description\":\"fip;tunnel_local-cidr\", \"type\":\"string\"}, \
        \"Subnet\":{\"description\":\"Subnetidonwhichvpnserviceislaunched\
        \", \"type\":\"string\"}}, \"resources\":{\"IKEPolicy\":{\"properties\
        \":{\"auth_algorithm\":\"sha1\", \"encryption_algorithm\":\"3des\", \
        \"ike_version\":\"v1\", \"lifetime\":{\"units\":\"seconds\", \"value\
        \":3600}, \"name\":\"IKEPolicy\", \"pfs\":\"group5\", \
        \"phase1_negotiation_mode\":\"main\"}, \"type\":\
        \"OS::Neutron::IKEPolicy\"}, \"IPsecPolicy\":{\"properties\":{\
        \"auth_algorithm\":\"sha1\", \"encapsulation_mode\":\"tunnel\", \
        \"encryption_algorithm\":\"3des\", \"lifetime\":{\"units\":\"seconds\
        \", \"value\":3600}, \"name\":\"IPsecPolicy\", \"pfs\":\"group5\", \
        \"transform_protocol\":\"esp\"}, \"type\":\"OS::Neutron::IPsecPolicy\
        \"}, \"VPNService\":{\"properties\":{\"admin_state_up\":\"true\", \
        \"description\":{\"get_param\":\"ServiceDescription\"}, \"name\":\
        \"VPNService\", \"router_id\":{\"get_param\":\"RouterId\"}, \
        \"subnet_id\":{\"get_param\":\"Subnet\"}}, \"type\":\
        \"OS::Neutron::VPNService\"}, \"site_to_site_connection1\
        \":{\"properties\":{\"admin_state_up\":\"true\", \"dpd\":{\"actions\
        \":\"hold\", \"interval\":30, \"timeout\":120}, \"ikepolicy_id\":{\
        \"get_resource\":\"IKEPolicy\"}, \"initiator\":\"bi-directional\", \
        \"ipsecpolicy_id\":{\"get_resource\":\"IPsecPolicy\"}, \"mtu\":1500, \
        \"name\":\"site_to_site_connection1\", \"peer_address\":\
        \"192.168.102.117\", \"peer_cidrs\":[\"11.0.0.0/24\"], \"peer_id\":\
        \"11.0.0.3\", \"psk\":\"secret\", \"vpnservice_id\":{\"get_resource\
        \":\"VPNService\"}}, \"type\":\"OS::Neutron::IPsecSiteConnection\"}}}"

    service_profile = {
        'service_flavor': 'vyos',
        'service_type': 'FIREWALL'
    }

    vpn_service_profile = {
        'service_flavor': 'vyos',
        'service_type': 'VPN'
    }

    lbv2_service_profile = {
        'service_flavor': 'haproxy',
        'service_type': 'LOADBALANCERV2'
    }

    fw_service_chain_node = {
        'id': '012345678919',
        'name': 'scn_fw',
        'config': fw_scn_config
    }

    vpn_service_chain_node = {
        'id': '012345678919',
        'name': 'scn_vpn',
        'config': vpn_scn_config
    }

    lbv2_service_chain_node = {
        'id': '012345678919',
        'name': 'scn_lb',
        'config': lbv2_scn_config
    }

    service_chain_instance = {
        'id': '7834569034456677',
        'name': 'sci_fw'
    }

    consumer_port = {
        'fixed_ips': [{
            'ip_address': '11.0.3.4',
            'subnet_id': '9876256378888333'
        }],
        'id': 'af6a8a58-1e25-49c4-97a3-d5f50b3aa04b'
    }

    network_function_details = {
        'network_function': {
            'status': 'ACTIVE',
            'description': '\nuser_access_ip=\'192.168.203.12\';'
                           'fixed_ip=\'11.0.3.4\';'
                           'tunnel_local_cidr=\'11.0.3.0/24\'',
            'config_policy_id': '57d6b523-ae89-41cd-9b63-9bfb054a20b6',
            'tenant_id': 'ee27b1d0d7f04ac390ee7ec4b2fd5b13',
            'network_function_instances': [
                '4693118c-149a-46e7-b92c-cc729b536a2e'],
            'service_chain_id': '507988d2-4b46-4df4-99d2-746676500872',
            'service_id': '1200332d-b432-403b-8350-89b782256be5',
            'service_profile_id': 'ab3b704b-a7d9-4c55-ab43-57ed5e29867d',
            'id': '5ad7439b-7259-47cd-be88-36f641e0b5c8',
            'name':
            'LOADBALANCERV2.haproxy.507988d2-4b46-4df4-99d2-746676500872'
        },
        'network_function_instance': {
            'status': 'ACTIVE',
            'name':
            'LOADBALANCERV2.haproxy.507988d2-4b46-4df4-99d2-7466765002',
            'network_function_device_id': '3c3e502a-256e-4597-91a9-71902380c0',
            'tenant_id': 'ee27b1d0d7f04ac390ee7ec4b2fd5b13',
            'ha_state': None,
            'network_function_id': '5ad7439b-7259-47cd-be88-36f641e0b5c8',
            'port_info': ['8cdcc00b-b791-4039-a5b4-e4d8b3d59e9f'],
            'id': '4693118c-149a-46e7-b92c-cc729b536a2e',
            'description': None
        },
        'network_function_device': {
            'status': 'ACTIVE',
            'monitoring_port_network': None,
            'monitoring_port_id': None,
            'mgmt_ip_address': '11.0.0.27',
            'description': '',
            'service_vendor': None,
            'tenant_id': 'ee27b1d0d7f04ac390ee7ec4b2fd5b13',
            'max_interfaces': 8,
            'mgmt_port_id': '4497a287-d947-4845-af29-a9d6ad6515e9',
            'reference_count': 1,
            'interfaces_in_use': 2,
            'id': '3c3e502a-256e-4597-91a9-719023808ec0',
            'name': 'LOADBALANCERV2.haproxy.507988d2-4b46-4df4-99d2-7466765008'
        }
    }

    _service_details = {
        'consuming_external_policies': [{
            'status': None,
            'consumed_policy_rule_sets': (
                                ['46de9c30-3f87-4fb7-8e56-5e60827e1e8f']),
            'external_segments': ['7648db78-f0e4-403d-91d4-c6d80963d56c'],
            'description': '',
            'tenant_id': '793827b52b3348929e97b23081dfac27',
            'provided_policy_rule_sets': [],
            'shared': False,
            'status_details': None,
            'id': 'aa06bb8b-1250-40e0-a1d0-e25a713cc978',
            'name': 'vpn-consumer'}],
        'service_vendor': 'vyos',
        'image_name': 'vyos',
        'network_mode': 'gbp',
        'consuming_ptgs_details': [],
        'device_type': 'nova',
        'service_type': 'VPN'
    }

    _subnet = {
        'shared': False,
        'description': None,
        'enable_dhcp': True,
        'network_id': '1e8612e0-8099-4577-ac27-97e7db6f5841',
        'tenant_id': '793827b52b3348929e97b23081dfac27',
        'created_at': '2016-07-26T17:05:11',
        'dns_nameservers': [],
        'updated_at': '2016-07-26T17:05:21',
        'ipv6_ra_mode': None,
        'allocation_pools': [{
                        'start': '192.168.0.2',
                        'end': '192.168.0.14'}],
        'gateway_ip': '192.168.0.1',
        'ipv6_address_mode': None,
        'ip_version': 4,
        'host_routes': [{
                    'nexthop': '192.168.0.3',
                    'destination': '12.0.0.0/24'}],
        'cidr': '192.168.0.0/28',
        'id': 'bab31ffb-07e1-42e9-a2b0-776efbf10f4a',
        'subnetpool_id': None,
        'name': 'ptg_tscp_1_vpn-provider'
    }

    service_details = {
        'service_details': _service_details,
        'provider_subnet': _subnet,
        'consumer_subnet': _subnet,
    }
    fip = [{'floating_ip_address': '192.168.102.118',
            'port_id': 'af6a8a58-1e25-49c4-97a3-d5f50b3aa04b'}]
    mgmt_ip = '11.3.4.5'
    l2p = {
           'l3_policy_id': '760d1763-9111-410a-a03e-61623afd7b25'
    }

    l3p = {
        'routers': ['64803e64-7db7-4050-a343-cbafbd2d356a']
           }
    services_nsp = [{'id': '479982d1-7947-478f-bf6c-dc234f38677d'}]
    stitching_pts = [{
            'policy_target_group_id': '6fa92b57-69ee-4143-9cf9-fcef0d067e65'
                }]



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# The following is imported at the beginning to ensure
# that the patches are applied before any of the
# modules save a reference to the functions being patched
from gbpservice.neutron.extensions import patch  # noqa

from neutron.common import utils as n_utils
from neutron.db.models import securitygroup as securitygroups_db
from neutron.db import models_v2
from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.plugins.ml2 import managers as ml2_managers
from neutron.plugins.ml2 import plugin as ml2_plugin
from neutron.quota import resource_registry
from neutron_lib.api.definitions import address_scope as as_def
from neutron_lib.api.definitions import network as net_def
from neutron_lib.api.definitions import port as port_def
from neutron_lib.api.definitions import subnet as subnet_def
from neutron_lib.api.definitions import subnetpool as subnetpool_def
from neutron_lib.api import validators
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib.db import resource_extend
from neutron_lib.db import utils as db_utils
from neutron_lib.plugins import directory
from oslo_log import log
from oslo_utils import excutils

from gbpservice.neutron.db import api as db_api
from gbpservice.neutron.db import implicitsubnetpool_db
from gbpservice.neutron.plugins.ml2plus import driver_api as api_plus
from gbpservice.neutron.plugins.ml2plus import driver_context
from gbpservice.neutron.plugins.ml2plus import managers

LOG = log.getLogger(__name__)


@registry.has_registry_receivers
@resource_extend.has_resource_extenders
class Ml2PlusPlugin(ml2_plugin.Ml2Plugin,
                    implicitsubnetpool_db.ImplicitSubnetpoolMixin):

    """Extend the ML2 core plugin with missing functionality.

    The standard ML2 core plugin in Neutron is missing a few features
    needed for optimal APIC AIM support. This class adds those
    features, while maintaining compatibility with all standard ML2
    drivers and configuration. The only change necessary to use
    ML2Plus is to register the ml2plus entry point instead of the ml2
    entry port as Neutron's core_plugin. Drivers that need these
    features inherit from the extended MechanismDriver and
    ExtensionDriver abstract base classes.
    """

    __native_bulk_support = True
    __native_pagination_support = True
    __native_sorting_support = True
    ml2_plugin.Ml2Plugin._supported_extension_aliases += [
        "implicit-subnetpools"]

    # Override and bypass immediate base class's __init__ in order to
    # instantate extended manager class(es).
    @resource_registry.tracked_resources(
        network=models_v2.Network,
        port=models_v2.Port,
        subnet=models_v2.Subnet,
        subnetpool=models_v2.SubnetPool,
        security_group=securitygroups_db.SecurityGroup,
        security_group_rule=securitygroups_db.SecurityGroupRule)
    def __init__(self):
        LOG.info("Ml2Plus initializing")
        # First load drivers, then initialize DB, then initialize drivers
        self.type_manager = ml2_managers.TypeManager()
        self.extension_manager = managers.ExtensionManager()
        self.mechanism_manager = managers.MechanismManager()
        super(ml2_plugin.Ml2Plugin, self).__init__()
        self.type_manager.initialize()
        self.extension_manager.initialize()
        self.mechanism_manager.initialize()
        self._setup_dhcp()
        self._start_rpc_notifiers()
        self.add_agent_status_check_worker(self.agent_health_check)
        self._verify_service_plugins_requirements()
        LOG.info("Modular L2 Plugin (extended) initialization complete")

    def start_rpc_listeners(self):
        servers = super(Ml2PlusPlugin, self).start_rpc_listeners()
        servers.extend(self.mechanism_manager.start_rpc_listeners())
        return servers

    # REVISIT: Handle directly in mechanism driver?
    @registry.receives(resources.SECURITY_GROUP,
                       [events.PRECOMMIT_CREATE, events.PRECOMMIT_UPDATE,
                        events.PRECOMMIT_DELETE])
    def _handle_security_group_change(self, resource, event, trigger,
                                      **kwargs):
        if 'payload' in kwargs:
            context = kwargs['payload'].context
            if event == events.PRECOMMIT_UPDATE:
                security_group = kwargs['payload'].desired_state
                original_security_group = kwargs['payload'].states[0]
            else:
                security_group = kwargs['payload'].states[0]
                original_security_group = kwargs['payload'].desired_state
        else:
            context = kwargs.get('context')
            security_group = kwargs.get('security_group')
            original_security_group = kwargs.get('original_security_group')
        # There is a neutron bug that sometimes it will create a SG with
        # tenant_id field empty. We will not process it further when that
        # happens then.
        if not security_group['tenant_id']:
            return
        mech_context = driver_context.SecurityGroupContext(
            self, context, security_group, original_security_group)
        if event == events.PRECOMMIT_CREATE:
            self._ensure_tenant(context, security_group)
            self.mechanism_manager.create_security_group_precommit(
                mech_context)
            return
        if event == events.PRECOMMIT_DELETE:
            self.mechanism_manager.delete_security_group_precommit(
                mech_context)
            return
        if event == events.PRECOMMIT_UPDATE:
            self.mechanism_manager.update_security_group_precommit(
                mech_context)

    # REVISIT: Handle directly in mechanism driver?
    @registry.receives(resources.SECURITY_GROUP_RULE,
                       [events.PRECOMMIT_CREATE, events.PRECOMMIT_DELETE])
    def _handle_security_group_rule_change(self, resource, event, trigger,
                                           **kwargs):
        if 'payload' in kwargs:
            context = kwargs['payload'].context
        else:
            context = kwargs.get('context')
        if event == events.PRECOMMIT_CREATE:
            if 'payload' in kwargs:
                sg_rule = kwargs['payload'].states[0]
            else:
                sg_rule = kwargs.get('security_group_rule')
            mech_context = driver_context.SecurityGroupRuleContext(
                self, context, sg_rule)
            self.mechanism_manager.create_security_group_rule_precommit(
                mech_context)
            return
        if event == events.PRECOMMIT_DELETE:
            if 'payload' in kwargs:
                sg_rule = {'id': kwargs['payload'].resource_id,
                           'security_group_id':
                               kwargs['payload'].metadata['security_group_id'],
                           'tenant_id': context.project_id}
            else:
                sg_rule = {'id': kwargs.get('security_group_rule_id'),
                        'security_group_id': kwargs.get('security_group_id'),
                        'tenant_id': context.project_id}
            mech_context = driver_context.SecurityGroupRuleContext(
                self, context, sg_rule)
            self.mechanism_manager.delete_security_group_rule_precommit(
                mech_context)

    @staticmethod
    @resource_extend.extends([net_def.COLLECTION_NAME])
    def _ml2_md_extend_network_dict(result, netdb):
        plugin = directory.get_plugin()
        session = db_api.get_session_from_obj(netdb)
        if session and session.is_active:
            # REVISIT: Check if transaction begin is still
            # required here, and if so, if reader pattern
            # can be used instead (will require getting the
            # current context, which should be available in
            # the session.info's dictionary, with a key of
            # 'using_context').
            with session.begin(subtransactions=True):
                plugin.extension_manager.extend_network_dict(
                        session, netdb, result)
        else:
            session = db_api.get_writer_session()
            plugin.extension_manager.extend_network_dict(
                    session, netdb, result)

    @staticmethod
    @resource_extend.extends([net_def.COLLECTION_NAME + '_BULK'])
    def _ml2_md_extend_network_dict_bulk(results, _):
        netdb = results[0][1] if results else None
        plugin = directory.get_plugin()
        session = db_api.get_session_from_obj(netdb)
        if session and session.is_active:
            with session.begin(subtransactions=True):
                plugin.extension_manager.extend_network_dict_bulk(session,
                                                                  results)
        else:
            session = db_api.get_writer_session()
            plugin.extension_manager.extend_network_dict_bulk(session, results)

    @staticmethod
    @resource_extend.extends([port_def.COLLECTION_NAME])
    def _ml2_md_extend_port_dict(result, portdb):
        plugin = directory.get_plugin()
        session = db_api.get_session_from_obj(portdb)
        if session and session.is_active:
            # REVISIT: Check if transaction begin is still
            # required here, and if so, if reader pattern
            # can be used instead (will require getting the
            # current context, which should be available in
            # the session.info's dictionary, with a key of
            # 'using_context').
            with session.begin(subtransactions=True):
                plugin.extension_manager.extend_port_dict(
                        session, portdb, result)
        else:
            session = db_api.get_writer_session()
            plugin.extension_manager.extend_port_dict(
                    session, portdb, result)

    @staticmethod
    @resource_extend.extends([port_def.COLLECTION_NAME + '_BULK'])
    def _ml2_md_extend_port_dict_bulk(results, _):
        portdb = results[0][1] if results else None
        plugin = directory.get_plugin()
        session = db_api.get_session_from_obj(portdb)
        if session and session.is_active:
            with session.begin(subtransactions=True):
                plugin.extension_manager.extend_port_dict_bulk(session,
                                                               results)
        else:
            session = db_api.get_writer_session()
            plugin.extension_manager.extend_port_dict_bulk(session, results)

    @staticmethod
    @resource_extend.extends([subnet_def.COLLECTION_NAME])
    def _ml2_md_extend_subnet_dict(result, subnetdb):
        plugin = directory.get_plugin()
        session = db_api.get_session_from_obj(subnetdb)
        if session and session.is_active:
            # REVISIT: Check if transaction begin is still
            # required here, and if so, if reader pattern
            # can be used instead (will require getting the
            # current context, which should be available in
            # the session.info's dictionary, with a key of
            # 'using_context').
            with session.begin(subtransactions=True):
                plugin.extension_manager.extend_subnet_dict(
                        session, subnetdb, result)
        else:
            session = db_api.get_writer_session()
            plugin.extension_manager.extend_subnet_dict(
                    session, subnetdb, result)

    @staticmethod
    @resource_extend.extends([subnet_def.COLLECTION_NAME + '_BULK'])
    def _ml2_md_extend_subnet_dict_bulk(results, _):
        subnetdb = results[0][1] if results else None
        plugin = directory.get_plugin()
        session = db_api.get_session_from_obj(subnetdb)
        if session and session.is_active:
            with session.begin(subtransactions=True):
                plugin.extension_manager.extend_subnet_dict_bulk(session,
                                                                 results)
        else:
            session = db_api.get_writer_session()
            plugin.extension_manager.extend_subnet_dict_bulk(session, results)

    @staticmethod
    @resource_extend.extends([subnetpool_def.COLLECTION_NAME])
    def _ml2_md_extend_subnetpool_dict(result, subnetpooldb):
        plugin = directory.get_plugin()
        session = db_api.get_session_from_obj(subnetpooldb)
        if session and session.is_active:
            # REVISIT: Check if transaction begin is still
            # required here, and if so, if reader pattern
            # can be used instead (will require getting the
            # current context, which should be available in
            # the session.info's dictionary, with a key of
            # 'using_context').
            with session.begin(subtransactions=True):
                plugin.extension_manager.extend_subnetpool_dict(
                        session, subnetpooldb, result)
        else:
            session = db_api.get_writer_session()
            plugin.extension_manager.extend_subnetpool_dict(
                    session, subnetpooldb, result)

    @staticmethod
    @resource_extend.extends([subnetpool_def.COLLECTION_NAME + '_BULK'])
    def _ml2_md_extend_subnetpool_dict_bulk(results, _):
        subnetpooldb = results[0][1] if results else None
        plugin = directory.get_plugin()
        session = db_api.get_session_from_obj(subnetpooldb)
        if session and session.is_active:
            with session.begin(subtransactions=True):
                plugin.extension_manager.extend_subnetpool_dict_bulk(session,
                                                                     results)
        else:
            session = db_api.get_writer_session()
            plugin.extension_manager.extend_subnetpool_dict_bulk(session,
                                                                 results)

    @staticmethod
    @resource_extend.extends([as_def.COLLECTION_NAME])
    def _ml2_md_extend_address_scope_dict(result, address_scope):
        plugin = directory.get_plugin()
        session = db_api.get_session_from_obj(address_scope)
        if session and session.is_active:
            # REVISIT: Check if transaction begin is still
            # required here, and if so, if reader pattern
            # can be used instead (will require getting the
            # current context, which should be available in
            # the session.info's dictionary, with a key of
            # 'using_context').
            with session.begin(subtransactions=True):
                plugin.extension_manager.extend_address_scope_dict(
                        session, address_scope, result)
        else:
            session = db_api.get_writer_session()
            plugin.extension_manager.extend_address_scope_dict(
                    session, address_scope, result)

    @staticmethod
    @resource_extend.extends([as_def.COLLECTION_NAME + '_BULK'])
    def _ml2_md_extend_address_scope_dict_bulk(results, _):
        address_scope = results[0][1] if results else None
        plugin = directory.get_plugin()
        session = db_api.get_session_from_obj(address_scope)
        if session and session.is_active:
            with session.begin(subtransactions=True):
                plugin.extension_manager.extend_address_scope_dict_bulk(
                    session, results)
        else:
            session = db_api.get_writer_session()
            plugin.extension_manager.extend_address_scope_dict_bulk(session,
                                                                    results)

    # Base version does not call _apply_dict_extend_functions()
    def _make_address_scope_dict(self, address_scope, fields=None):
        res = {'id': address_scope['id'],
               'name': address_scope['name'],
               'tenant_id': address_scope['tenant_id'],
               'shared': address_scope['shared'],
               'ip_version': address_scope['ip_version']}
        resource_extend.apply_funcs(
            as_def.COLLECTION_NAME, res, address_scope)
        return db_utils.resource_fields(res, fields)

    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_network(self, context, network):
        self._ensure_tenant(context, network[net_def.RESOURCE_NAME])
        return super(Ml2PlusPlugin, self).create_network(context, network)

    # We are overriding _create_network_db here to get
    # around a bug that was introduced in the following commit:
    # https://github.com/openstack/neutron/commit/
    # 2b7c6b2e987466973d983902eded6aff7f764830#
    # diff-2e958ca8f1a6e9987e28a7d0f95bc3d1L776
    # which moves the call to extending the dict before the call to
    # pre_commit. We need to extend_dict function to pick up the changes
    # from the pre_commit operations as well.
    def _create_network_db(self, context, network):
        with db_api.CONTEXT_WRITER.using(context):
            result, mech_context = super(
                    Ml2PlusPlugin, self)._create_network_db(
                            context, network)
            net_db = (context.session.query(models_v2.Network).
                      filter_by(id=result['id']).one())
            resource_extend.apply_funcs('networks', result, net_db)
            return result, mech_context

    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_network_bulk(self, context, networks):
        self._ensure_tenant_bulk(context, networks[net_def.COLLECTION_NAME],
                net_def.RESOURCE_NAME)
        return super(Ml2PlusPlugin, self).create_network_bulk(context,
                                                              networks)

    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_subnet(self, context, subnet):
        self._ensure_tenant(context, subnet[subnet_def.RESOURCE_NAME])
        return super(Ml2PlusPlugin, self).create_subnet(context, subnet)

    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_subnet_bulk(self, context, subnets):
        self._ensure_tenant_bulk(context, subnets[subnet_def.COLLECTION_NAME],
                                 subnet_def.RESOURCE_NAME)
        return super(Ml2PlusPlugin, self).create_subnet_bulk(context,
                                                             subnets)

    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_port(self, context, port):
        self._ensure_tenant(context, port[port_def.RESOURCE_NAME])
        return super(Ml2PlusPlugin, self).create_port(context, port)

    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_port_bulk(self, context, ports):
        self._ensure_tenant_bulk(context, ports[port_def.COLLECTION_NAME],
                                 port_def.RESOURCE_NAME)
        return super(Ml2PlusPlugin, self).create_port_bulk(context,
                                                           ports)

    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_subnetpool(self, context, subnetpool):
        self._ensure_tenant(context, subnetpool[subnetpool_def.RESOURCE_NAME])
        with db_api.CONTEXT_WRITER.using(context):
            result = super(Ml2PlusPlugin, self).create_subnetpool(context,
                                                                  subnetpool)
            self._update_implicit_subnetpool(context, subnetpool, result)
            self.extension_manager.process_create_subnetpool(
                context, subnetpool[subnetpool_def.RESOURCE_NAME], result)
            mech_context = driver_context.SubnetPoolContext(
                self, context, result)
            self.mechanism_manager.create_subnetpool_precommit(mech_context)
        try:
            self.mechanism_manager.create_subnetpool_postcommit(mech_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error("mechanism_manager.create_subnetpool_postcommit "
                          "failed, deleting subnetpool '%s'",
                          result['id'])
                self.delete_subnetpool(context, result['id'])
        return result

    # REVISIT(rkukura): Is create_subnetpool_bulk() needed?

    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_subnetpool(self, context, id, subnetpool):
        with db_api.CONTEXT_WRITER.using(context):
            original_subnetpool = super(Ml2PlusPlugin, self).get_subnetpool(
                context, id)
            updated_subnetpool = super(Ml2PlusPlugin, self).update_subnetpool(
                context, id, subnetpool)
            self._update_implicit_subnetpool(context, subnetpool,
                                             updated_subnetpool)
            self.extension_manager.process_update_subnetpool(
                context, subnetpool[subnetpool_def.RESOURCE_NAME],
                updated_subnetpool)
            mech_context = driver_context.SubnetPoolContext(
                self, context, updated_subnetpool,
                original_subnetpool=original_subnetpool)
            self.mechanism_manager.update_subnetpool_precommit(mech_context)
        self.mechanism_manager.update_subnetpool_postcommit(mech_context)
        return updated_subnetpool

    @n_utils.transaction_guard
    def delete_subnetpool(self, context, id):
        with db_api.CONTEXT_WRITER.using(context):
            subnetpool = super(Ml2PlusPlugin, self).get_subnetpool(context, id)
            mech_context = driver_context.SubnetPoolContext(
                self, context, subnetpool)
            self.mechanism_manager.delete_subnetpool_precommit(mech_context)
            super(Ml2PlusPlugin, self).delete_subnetpool(context, id)
        self.mechanism_manager.delete_subnetpool_postcommit(mech_context)

    def _update_implicit_subnetpool(self, context, request, result):
        if validators.is_attr_set(request['subnetpool'].get('is_implicit')):
            result['is_implicit'] = request['subnetpool']['is_implicit']
            result['is_implicit'] = (
                self.update_implicit_subnetpool(context, result))

    @n_utils.transaction_guard
    def create_address_scope(self, context, address_scope):
        self._ensure_tenant(context, address_scope[as_def.ADDRESS_SCOPE])
        with db_api.CONTEXT_WRITER.using(context):
            result = super(Ml2PlusPlugin, self).create_address_scope(
                context, address_scope)
            self.extension_manager.process_create_address_scope(
                context, address_scope[as_def.ADDRESS_SCOPE], result)
            mech_context = driver_context.AddressScopeContext(
                self, context, result)
            self.mechanism_manager.create_address_scope_precommit(
                mech_context)
        try:
            self.mechanism_manager.create_address_scope_postcommit(
                mech_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error("mechanism_manager.create_address_scope_"
                          "postcommit failed, deleting address_scope"
                          " '%s'",
                          result['id'])
                self.delete_address_scope(context, result['id'])
        return result

    # REVISIT(rkukura): Is create_address_scope_bulk() needed?

    @n_utils.transaction_guard
    def update_address_scope(self, context, id, address_scope):
        with db_api.CONTEXT_WRITER.using(context):
            original_address_scope = super(Ml2PlusPlugin,
                                           self).get_address_scope(context, id)
            updated_address_scope = super(Ml2PlusPlugin,
                                          self).update_address_scope(
                                              context, id, address_scope)
            self.extension_manager.process_update_address_scope(
                context, address_scope[as_def.ADDRESS_SCOPE],
                updated_address_scope)
            mech_context = driver_context.AddressScopeContext(
                self, context, updated_address_scope,
                original_address_scope=original_address_scope)
            self.mechanism_manager.update_address_scope_precommit(mech_context)
        self.mechanism_manager.update_address_scope_postcommit(mech_context)
        return updated_address_scope

    @n_utils.transaction_guard
    def delete_address_scope(self, context, id):
        with db_api.CONTEXT_WRITER.using(context):
            address_scope = super(Ml2PlusPlugin, self).get_address_scope(
                context, id)
            mech_context = driver_context.AddressScopeContext(
                self, context, address_scope)
            self.mechanism_manager.delete_address_scope_precommit(mech_context)
            super(Ml2PlusPlugin, self).delete_address_scope(context, id)
        self.mechanism_manager.delete_address_scope_postcommit(mech_context)

    def _ensure_tenant(self, context, resource):
        tenant_id = resource['tenant_id']
        self.mechanism_manager.ensure_tenant(context, tenant_id)

    def _ensure_tenant_bulk(self, context, resources, singular):
        tenant_ids = [resource[singular]['tenant_id']
                      for resource in resources]
        for tenant_id in set(tenant_ids):
            self.mechanism_manager.ensure_tenant(context, tenant_id)

    def _get_subnetpool_id(self, context, subnet):
        # Check for regular subnetpool ID first, then Tenant's implicit,
        # then global implicit.
        ip_version = subnet['ip_version']
        return (
            super(Ml2PlusPlugin, self)._get_subnetpool_id(context, subnet) or
            self.get_implicit_subnetpool_id(context,
                                            tenant=subnet['tenant_id'],
                                            ip_version=ip_version) or
            self.get_implicit_subnetpool_id(context, tenant=None,
                                            ip_version=ip_version))

    # REVISIT(ivar): patching bulk gets for extension performance

    def _make_networks_dict(self, networks, context):
        nets = []
        for network in networks:
            if network.mtu is None:
                # TODO(ivar): also refactor this to run for bulk networks
                network.mtu = self._get_network_mtu(network, validate=False)
            res = {'id': network['id'],
                   'name': network['name'],
                   'tenant_id': network['tenant_id'],
                   'admin_state_up': network['admin_state_up'],
                   'mtu': network.get('mtu', n_const.DEFAULT_NETWORK_MTU),
                   'status': network['status'],
                   'subnets': [subnet['id']
                               for subnet in network['subnets']]}
            res['shared'] = self._is_network_shared(context,
                                                    network.rbac_entries)
            nets.append((res, network))

        # Bulk extend first
        resource_extend.apply_funcs(net_def.COLLECTION_NAME + '_BULK', nets,
                                    None)

        result = []
        for res, network in nets:
            res[api_plus.BULK_EXTENDED] = True
            resource_extend.apply_funcs(net_def.COLLECTION_NAME, res, network)
            res.pop(api_plus.BULK_EXTENDED, None)
            result.append(db_utils.resource_fields(res, []))
        return result

    @db_api.retry_if_session_inactive()
    def get_networks(self, context, filters=None, fields=None,
                     sorts=None, limit=None, marker=None, page_reverse=False):
        with db_api.CONTEXT_WRITER.using(context):
            nets_db = super(Ml2PlusPlugin, self)._get_networks(
                context, filters, None, sorts, limit, marker, page_reverse)

            net_data = self._make_networks_dict(nets_db, context)

            self.type_manager.extend_networks_dict_provider(context, net_data)
            nets = self._filter_nets_provider(context, net_data, filters)
        return [db_utils.resource_fields(net, fields) for net in nets]

    def _make_subnets_dict(self, subnets_db, fields=None, context=None):
        subnets = []
        for subnet_db in subnets_db:
            res = {'id': subnet_db['id'],
                   'name': subnet_db['name'],
                   'tenant_id': subnet_db['tenant_id'],
                   'network_id': subnet_db['network_id'],
                   'ip_version': subnet_db['ip_version'],
                   'subnetpool_id': subnet_db['subnetpool_id'],
                   'enable_dhcp': subnet_db['enable_dhcp'],
                   'ipv6_ra_mode': subnet_db['ipv6_ra_mode'],
                   'ipv6_address_mode': subnet_db['ipv6_address_mode'],
                   }
            res['gateway_ip'] = str(
                    subnet_db['gateway_ip']) if subnet_db['gateway_ip'] else (
                    None)
            res['cidr'] = subnet_db['cidr']
            res['allocation_pools'] = [{'start': pool['first_ip'],
                                       'end': pool['last_ip']}
                                       for pool in
                                       subnet_db['allocation_pools']]
            res['host_routes'] = [{'destination': route['destination'],
                                  'nexthop': route['nexthop']}
                                  for route in subnet_db['routes']]
            res['dns_nameservers'] = [dns['address']
                                      for dns in subnet_db['dns_nameservers']]

            # The shared attribute for a subnet is the same
            # as its parent network
            res['shared'] = self._is_network_shared(context,
                                                    subnet_db.rbac_entries)

            subnets.append((res, subnet_db))

        resource_extend.apply_funcs(subnet_def.COLLECTION_NAME + '_BULK',
                                    subnets, None)

        result = []
        for res, subnet_db in subnets:
            res[api_plus.BULK_EXTENDED] = True
            resource_extend.apply_funcs(subnet_def.COLLECTION_NAME,
                                        res, subnet_db)
            res.pop(api_plus.BULK_EXTENDED, None)
            result.append(db_utils.resource_fields(res, []))

        return result

    @db_api.retry_if_session_inactive()
    def get_subnets(self, context, filters=None, fields=None,
                    sorts=None, limit=None, marker=None,
                    page_reverse=False):

        with db_api.CONTEXT_READER.using(context):
            plugin = directory.get_plugin()
            marker_obj = db_api.get_marker_obj(plugin, context,
                                               'subnet', limit, marker)

            # REVIST(sridar): We need to rethink if we want to support
            # OVO. For now we put our head in the sand but this needs a
            # revisit.
            # Also, older branches are a slight variation, in line with
            # upstream code.
            subnets_db = db_api.get_collection(context, models_v2.Subnet,
                                               dict_func=None,
                                               filters=filters,
                                               sorts=sorts,
                                               limit=limit,
                                               marker_obj=marker_obj,
                                               page_reverse=page_reverse)

            subnets = self._make_subnets_dict(subnets_db, fields, context)
        return [db_api.resource_fields(subnet, fields) for subnet in subnets]



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Group Policy Mapping DB (gpm_db_1)

Revision ID: 64fa77aca090

"""

# revision identifiers, used by Alembic.
revision = '64fa77aca090'
down_revision = 'f16efdc10a71'


from alembic import op
import sqlalchemy as sa


def upgrade():

    op.add_column(
        'gp_external_segments',
        sa.Column('subnet_id', sa.String(length=36), nullable=True)
    )
    op.add_column(
        'gp_external_segments',
        sa.Column('type', sa.String(length=15), nullable=True)
    )
    op.create_unique_constraint(None, 'gp_external_segments', ['subnet_id'])
    op.create_foreign_key('gp_external_segment_ibfk_2',
                          source_table='gp_external_segments',
                          referent_table='subnets',
                          local_cols=['subnet_id'], remote_cols=['id'])


def downgrade():

    op.drop_constraint('gp_external_segment_ibfk_2', 'gp_external_segments',
                       'foreignkey')
    op.drop_column('gp_external_segments', 'type')
    op.drop_column('gp_external_segments', 'subnet_id')



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.api import extensions
from neutron.db.models import address_scope as as_db
from neutron.db import models_v2
from neutron.objects import address_scope as as_object
from neutron_lib.api import validators
from neutron_lib.db import model_base
import oslo_db.sqlalchemy.session
import sqlalchemy as sa
from sqlalchemy import orm

from gbpservice.neutron.plugins.ml2plus import driver_api
from gbpservice.neutron.tests.unit.plugins.ml2plus import (
    extensions as test_extensions)


class TestExtensionDriverBase(driver_api.ExtensionDriver):
    _supported_extension_aliases = 'ml2plus_fake_extension'

    def initialize(self):
        # REVISIT(rkukura): Needed?
        extensions.append_api_extensions_path(test_extensions.__path__)

    @property
    def extension_alias(self):
        return self._supported_extension_aliases


class TestExtensionDriver(TestExtensionDriverBase):
    def initialize(self):
        super(TestExtensionDriver, self).initialize()
        self.subnetpool_extension = 'Test_SubnetPool_Extension'
        self.address_scope_extension = 'Test_AddressScope_Extension'

    def _check_create(self, session, data, result):
        assert(isinstance(session, oslo_db.sqlalchemy.session.Session))
        assert(isinstance(data, dict))
        assert('id' not in data)
        assert(isinstance(result, dict))
        assert(result['id'] is not None)

    def _check_update(self, session, data, result):
        assert(isinstance(session, oslo_db.sqlalchemy.session.Session))
        assert(isinstance(data, dict))
        assert(isinstance(result, dict))
        assert(result['id'] is not None)

    def _check_extend(self, session, result, db_entry,
                      expected_db_entry_class):
        assert(isinstance(session, oslo_db.sqlalchemy.session.Session))
        assert(isinstance(result, dict))
        assert(result['id'] is not None)
        assert(isinstance(db_entry, expected_db_entry_class))
        assert(db_entry.id == result['id'])

    def process_create_subnetpool(self, plugin_context, data, result):
        session = plugin_context.session
        self._check_create(session, data, result)
        result['subnetpool_extension'] = self.subnetpool_extension + '_create'

    def process_update_subnetpool(self, plugin_context, data, result):
        session = plugin_context.session
        self._check_update(session, data, result)
        self.subnetpool_extension = data['subnetpool_extension']
        result['subnetpool_extension'] = self.subnetpool_extension + '_update'

    def extend_subnetpool_dict(self, session, subnetpool_db, result):
        self._check_extend(session, result, subnetpool_db,
                           models_v2.SubnetPool)
        result['subnetpool_extension'] = self.subnetpool_extension + '_extend'

    def process_create_address_scope(self, plugin_context, data, result):
        session = plugin_context.session
        self._check_create(session, data, result)
        result['address_scope_extension'] = (self.address_scope_extension +
                                             '_create')

    def process_update_address_scope(self, plugin_context, data, result):
        session = plugin_context.session
        self._check_update(session, data, result)
        self.address_scope_extension = data['address_scope_extension']
        result['address_scope_extension'] = (self.address_scope_extension +
                                             '_update')

    def extend_address_scope_dict(self, session, address_scope, result):
        self._check_extend(session, result, address_scope,
                           as_object.AddressScope)
        result['address_scope_extension'] = (self.address_scope_extension +
                                             '_extend')


class TestSubnetPoolExtension(model_base.BASEV2):
    subnetpool_id = sa.Column(sa.String(36),
                              sa.ForeignKey('subnetpools.id',
                                            ondelete="CASCADE"),
                              primary_key=True)
    value = sa.Column(sa.String(64))
    subnetpool = orm.relationship(
        models_v2.SubnetPool,
        backref=orm.backref('extension', cascade='delete', uselist=False))


class TestAddressScopeExtension(model_base.BASEV2):
    address_scope_id = sa.Column(sa.String(36),
                                 sa.ForeignKey('address_scopes.id',
                                               ondelete="CASCADE"),
                                 primary_key=True)
    value = sa.Column(sa.String(64))
    address_scope = orm.relationship(
        as_db.AddressScope,
        backref=orm.backref('extension', cascade='delete', uselist=False))


class TestDBExtensionDriver(TestExtensionDriverBase):
    def _get_value(self, data, key):
        value = data[key]
        if not validators.is_attr_set(value):
            value = ''
        return value

    def process_create_subnetpool(self, plugin_context, data, result):
        session = plugin_context.session
        value = self._get_value(data, 'subnetpool_extension')
        record = TestSubnetPoolExtension(subnetpool_id=result['id'],
                                         value=value)
        session.add(record)
        result['subnetpool_extension'] = value

    def process_update_subnetpool(self, plugin_context, data, result):
        session = plugin_context.session
        record = (session.query(TestSubnetPoolExtension).
                  filter_by(subnetpool_id=result['id']).one())
        value = data.get('subnetpool_extension')
        if value and value != record.value:
            record.value = value
        result['subnetpool_extension'] = record.value

    def extend_subnetpool_dict(self, session, subnetpool_db, result):
        record = (session.query(TestSubnetPoolExtension).
                 filter_by(subnetpool_id=result['id']).one_or_none())
        result['subnetpool_extension'] = record.value if record else ''

    def process_create_address_scope(self, plugin_context, data, result):
        session = plugin_context.session
        value = self._get_value(data, 'address_scope_extension')
        record = TestAddressScopeExtension(address_scope_id=result['id'],
                                           value=value)
        session.add(record)
        result['address_scope_extension'] = value

    def process_update_address_scope(self, plugin_context, data, result):
        session = plugin_context.session
        record = (session.query(TestAddressScopeExtension).
                  filter_by(address_scope_id=result['id']).one())
        value = data.get('address_scope_extension')
        if value and value != record.value:
            record.value = value
        result['address_scope_extension'] = record.value

    def extend_address_scope_dict(self, session, address_scope, result):
        record = (session.query(TestAddressScopeExtension).
                 filter_by(address_scope_id=result['id']).one_or_none())
        result['address_scope_extension'] = record.value if record else ''



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from oslo_log import log as logging

from gbpservice._i18n import _
from gbpservice.neutron.db.grouppolicy.extensions import (
    apic_allowed_vm_name_db as db)
from gbpservice.neutron.extensions import apic_allowed_vm_name as aavnext
from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)
from gbpservice.neutron.services.grouppolicy.common import exceptions as gpexc

LOG = logging.getLogger(__name__)


class AllowedVMNameBadRegex(gpexc.GroupPolicyBadRequest):
    message = _("Bad regex: %(regex)s is defined for the allowed-vm-names "
                "attribute.")


class ApicAllowedVMNameExtensionDriver(api.ExtensionDriver,
                                       db.ApicAllowedVMNameDBMixin):
    _supported_extension_alias = aavnext.CISCO_APIC_GBP_ALLOWED_VM_NAME_EXT
    _extension_dict = aavnext.EXTENDED_ATTRIBUTES_2_0

    def __init__(self):
        LOG.debug("APIC Allowed VM Name Extension Driver  __init__")
        self._policy_driver = None

    def initialize(self):
        pass

    @property
    def extension_alias(self):
        return self._supported_extension_alias

    def process_create_l3_policy(self, session, data, result):
        l3p = data['l3_policy']
        if 'allowed_vm_names' in l3p:
            for vm_name in l3p['allowed_vm_names']:
                try:
                    re.compile(vm_name)
                except re.error:
                    raise AllowedVMNameBadRegex(regex=vm_name)
                self.add_l3_policy_allowed_vm_name(
                    session, l3_policy_id=result['id'],
                    allowed_vm_name=vm_name)

    def process_update_l3_policy(self, session, data, result):
        l3p = data['l3_policy']
        if 'allowed_vm_names' not in l3p:
            return
        rows = self.get_l3_policy_allowed_vm_names(
            session, l3_policy_id=result['id'])
        old_vm_names = [r.allowed_vm_name for r in rows]
        add_vm_names = list(set(l3p['allowed_vm_names']) - set(old_vm_names))
        for vm_name in add_vm_names:
            try:
                re.compile(vm_name)
            except re.error:
                raise AllowedVMNameBadRegex(regex=vm_name)
            self.add_l3_policy_allowed_vm_name(
                session, l3_policy_id=result['id'],
                allowed_vm_name=vm_name)
        rm_vm_names = list(set(old_vm_names) - set(l3p['allowed_vm_names']))
        for vm_name in rm_vm_names:
            self.delete_l3_policy_allowed_vm_name(
                session, l3_policy_id=result['id'],
                allowed_vm_name=vm_name)

    def extend_l3_policy_dict(self, session, result):
        rows = self.get_l3_policy_allowed_vm_names(
            session, l3_policy_id=result['id'])
        allowed_vm_names = [r.allowed_vm_name for r in rows]
        result['allowed_vm_names'] = allowed_vm_names
